#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class RedborderObjectsController < ApplicationController
  # -------------------------------------------------------------
  # Before actions
  # -------------------------------------------------------------

  before_action :set_object, only: [:edit, :update, :destroy]
  before_action :set_domain_names, only: [:index, :edit, :update, :create, :search]

  # -------------------------------------------------------------
  # Actions
  # -------------------------------------------------------------

  #
  # Render all the objects inside tabs.
  #
  def index
    session.delete(:search)
    
    sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

    @net_objects = fetch_objects(NetObject, sensor_ids)
                    .order(created_at: :desc)
                    .paginate(page: 1, per_page: current_user.per_page_count)
  
    @mac_objects = fetch_objects(MacObject, sensor_ids)
                    .order(created_at: :desc)
                    .paginate(page: 1, per_page: current_user.per_page_count)

    inventory_ids = @mac_objects.map(&:inventory_device_id).compact.uniq
    @inventory_devices = InventoryDevice.where(id: inventory_ids).index_by(&:id)
  
    @vlan_objects       = current_user
                          .vlan_objects
                          .order(created_at: :desc)
                          .paginate(page: 1, per_page: current_user.per_page_count)
    @service_objects    = current_user
                          .service_objects
                          .order(created_at: :desc)
                          .paginate(page: 1, per_page: current_user.per_page_count)
    @application_objects = ApplicationObject.all
                            .order(created_at: :desc)
                            .paginate(page: 1, per_page: current_user.per_page_count)
    @operating_system_objects = OperatingSystemObject.all
                            .order(created_at: :desc)
                            .paginate(page: 1, per_page: current_user.per_page_count)
    @hardware_objects = HardwareObject.all
                            .order(created_at: :desc)
                            .paginate(page: 1, per_page: current_user.per_page_count)
    @inventory_device_type_objects = InventoryDeviceTypeObject.all
                            .order(created_at: :desc)
                            .paginate(page: 1, per_page: current_user.per_page_count)
  end

  #
  # Render the form to create a new object
  #
  def new
    return unless RedborderObject::OBJECT_CLASSES.include? params[:class]
    @object = params[:class].constantize.new(value: params[:value])
    render layout: false
  end

  #
  # Render the form to edit an object
  #
  def edit
    # Define a whitelist of allowed object types
    allowed_classes = {
      'NetObject' => NetObject,
      'MacObject' => MacObject,
      'ProtocolObject' => ProtocolObject,
      'ServiceObject' => ServiceObject,
      'VlanObject' => VlanObject
    }

    object_type_param = params[:object][:type]

    # Validate that the object type is in the whitelist
    unless allowed_classes.key?(object_type_param)
      rb_flash "#{I18n.t('controllers.object_not_found')}", type: :error
      redirect_to objects_path and return
    end

    # Find the object based on the class and ID
    object_type = allowed_classes[object_type_param]
    @object = object_type.find_by(id: params[:id])

    unless @object
      rb_flash "#{I18n.t('controllers.object_not_found')}", type: :error
      redirect_to objects_path and return
    end

    render layout: false
  end

  #
  # Receive the POST data of an object and try to save it
  #
  def create
    # Create the object with the received parameters
    @object = RedborderObject.new(object_params)

    # Assign user_id and sensor_id
    assign_user_and_sensor

    # Get all objects to calculate new pagination
    if @object.save
      sensor_ids = current_user.domains_with_manage_permission.pluck(:id)
      
      if session[:search].present?
        filter = ['( LOWER(name) LIKE :search or ' \
                'LOWER(comment) LIKE :search or LOWER(value) LIKE :search)',
                search: "%#{session[:search].try(:downcase)}%"]
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids).where(filter)
      else
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids)
      end

      new_page = params[:page] ? params[:page].to_i : 1
      @objects = @objects.order(created_at: :desc).paginate(page: new_page, per_page: current_user.per_page_count)
      flash.now[:notice] = I18n.t('redborder_objects.create.success')
    else
      flash[:alert] = I18n.t('redborder_objects.create.error')
    end
  end

  #
  # Receive the POST data of an object and try to update it
  #
  def update
    # Get all domain IDs that the current user can manage
    assign_user_and_sensor

    # Attempt to update the object with the received parameters (excluding the sensor)
    if @object.update(object_params.except(:sensor))
      flash.now[:notice] = I18n.t('redborder_objects.update.success')
      render layout: false
    else
      flash[:alert] = I18n.t('redborder_objects.update.error')
      render :edit
    end
  end

  #
  # Destroy an object
  #
  def destroy
    if @object && @object.destroy
      sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

      # Keep filter when deleting
      if session[:search].present?
        filter = ['( LOWER(name) LIKE :search or ' \
                'LOWER(comment) LIKE :search or LOWER(value) LIKE :search)',
                search: "%#{session[:search].try(:downcase)}%"]
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids).where(filter)
      else
        @objects = fetch_objects(object_params[:type].classify.constantize, sensor_ids)
      end

      # Keep page in range and deleting last page if it is empty
      current_page = [params[:page].to_i, 1].max
      max_page = (@objects.any? ? (@objects.count.to_f / current_user.per_page_count).ceil : 1)
      new_page = [current_page, max_page].min

      # Show table fixed (pagination)
      @objects = @objects.order(created_at: :desc).paginate(page: new_page, per_page: current_user.per_page_count)
      flash.now[:notice] = I18n.t('redborder_objects.destroy.success')
    else
      flash[:alert] = I18n.t('redborder_objects.destroy.error')
    end
  end

  #
  # Store session tab of the user
  #
  def change_session_tab
    session[:object] = params[:id].to_sym
    render js: ''
  end

  # TODO: THIS NEED TO BE DELETED
  # def apply
  #   @sensor.apply_objects
  #   Manager.apply_objects
  #   respond_to do |format|
  #     format.html { redirect_to :back, notice: 'Objects has been applied.' }
  #     format.json { head :no_content }
  #   end
  # end

  #
  # Search in objects
  #
  def search
    begin
      @object_type = RedborderObject.from_string(params[:tab])
    rescue RedborderObject::Unknown
      error_400("#{I18n.t('controllers.no_object')}") && return
    end
    sensor_ids = current_user.domains_with_manage_permission.pluck(:id)

    if params[:search].blank?
      session.delete(:search)
      @objects = fetch_objects(@object_type.classify.constantize, sensor_ids)
              .order(created_at: :desc)
              .paginate(page: params[:page], per_page: current_user.per_page_count)
    else
      session[:search] = params[:search] if params[:search].present?
      filter = ['( LOWER(name) LIKE :search or ' \
              'LOWER(comment) LIKE :search or LOWER(value) LIKE :search or LOWER(vendor) LIKE :search )',
              search: "%#{session[:search].try(:downcase)}%"]
      @objects = fetch_objects(@object_type.classify.constantize, sensor_ids)
              .where(filter)
              .order(created_at: :desc)
              .paginate(page: params[:page], per_page: current_user.per_page_count)
    end

    if @object_type == 'MacObject'
      inventory_ids = @objects.map(&:inventory_device_id).compact.uniq
      @inventory_devices = InventoryDevice.where(id: inventory_ids).index_by(&:id)
    end
  end

  #
  # Export all objects as CSS
  #
  def export
    tab = params[:tab]
    export_objects_csv tab
  end

  #
  # Show a modal to upload a file to import objects.
  #
  def import
    @type = params[:type]
    render layout: false
  end

  #
  # Import objects from a given CSV file.
  #
  def import_file
    tab = params[:tab]
    content_file = params[:redborder_object][:csv_file]

    unless content_file.content_type == 'text/csv'
      redirect_to objects_path, alert: t('controllers.unknow_content_csv')
      return
    end

    begin
      import_from_csv(content_file, current_user, tab)
      msg = t('controllers.objects_added')
      redirect_to objects_path, notice: msg
    rescue => e
      msg = t('controllers.objects_added_no_valid', error: e.message)
      redirect_to objects_path, alert: msg
    end
  end

  

  private

  # This function retrieves objects of a given class based on the following conditions:
  # 1. Returns objects belonging to the current user (user_id = current_user.id).
  # 2. If sensor IDs are provided, includes objects where user_id is NULL and sensor_id is in the list of allowed IDs.
  # 3. If the current user is a super admin, also includes objects where sensor_id is NULL.
  # It returns an ActiveRecord::Relation with all objects that meet these conditions.
  def fetch_objects(object_class, sensor_ids = [])

    conditions = object_class.where(user_id: current_user.id)
  
    if sensor_ids.present?
      conditions = conditions.or(object_class.where(user_id: nil, sensor_id: sensor_ids))
    end
  
    if current_user.super_admin?
      conditions = conditions.or(object_class.where(sensor_id: nil))
    end
  
    conditions
  end  

  # Set @domain_names
  def set_domain_names
    @domain_names = current_user.domains_with_manage_permission.map { |d| [d.id, d.path_name] }.to_h
  end

  # Method to assign user_id and sensor_id to the object based on the provided parameters.
  def assign_user_and_sensor
    # If the public checkox is checked or if the asset has a inventory device associated
    if params[:check] == "1" || !@object.inventory_device.nil?
      @object.user_id = nil
      sensor = @object.inventory_device.nil? ? current_user.sensor_id : @object.inventory_device.sensor_id
      sensor = params[:sensor_id] unless params[:sensor_id].nil?
      @object.sensor_id = sensor
    else
      @object.user_id = current_user.id
      @object.sensor_id = current_user.sensor_id
    end
  end

  #
  # Set the object.
  #
  def set_object
    # Convert the type name to a class format
    type = RedborderObject.from_string(params[:object][:type])
    model_class = type.classify.constantize  # Convert the name to CamelCase format and then get the class

    @object = model_class
                .where(id: params[:id])
                .where("user_id = ? OR user_id IS NULL", current_user.id)
                .first

    return if @object

    unless @object
      rb_flash "#{I18n.t('controllers.object_not_exist')}", type: :error
      return
    end
  end

  #
  # Export objects as CSV
  #
  def export_objects_csv(tab)
    filename = in_time_machine { Time.now.strftime('%Y%m%d%H%M%S') +"-#{tab.capitalize}-Objects.csv" }
    content = objects_to_csv_format(tab)
    send_data content, filename: filename
  end

  #
  # Convert current Objects to CSV
  #
  def objects_to_csv_format(tab)
    columns = RedborderObject.column_names

    CSV.generate(col_sep: ';') do |csv|
      csv << columns

      search_term = params[:search].presence || ''
      search_term = '' if search_term == 'undefined'
      sensor_ids = current_user.domains_with_permission.pluck(:id)

      # Fetch objects based on user permissions
      base_scope = RedborderObject.where(type: "#{tab.capitalize}Object")
      objects_with_permissions = base_scope.where(user_id: current_user.id)

      if sensor_ids.present?
        objects_with_permissions = objects_with_permissions.or(base_scope.where(user_id: nil, sensor_id: sensor_ids))
      end

      if current_user.super_admin?
        objects_with_permissions = objects_with_permissions.or(base_scope.where(sensor_id: nil))
      end

      # Define the filter for the search term
      filter = ['(LOWER(name) LIKE :search OR LOWER(comment) LIKE :search OR LOWER(value) LIKE :search)',
      { search: "%#{search_term.downcase}%" }]

      # Generate CSV content
      objects_with_permissions.where(filter).order(:name).each do |object|
        csv << columns.map { |col| object.send(col) }
      end
    end
  end

  #
  # Import Objects from CSV
  #
  def import_from_csv(content_file, current_user, tab)
    csv_data = CSV.read(content_file.path, col_sep: ';', headers: true, encoding: 'UTF-8')
    columns = csv_data.headers.map(&:to_sym)

    CSV.foreach(content_file.path, col_sep: ';', headers: false, encoding: 'UTF-8') do |row|
      next if row.empty? || row[0].nil?

      attrs = {}
      columns.each_with_index do |col, index|
        attrs[col] = row[index].try(:strip)
      end

      next if (current_user.domains_with_permission.pluck(:id).include?(attrs[:sensor_id] || 1) || attrs[:type] != "#{tab.capitalize}Object")

      attrs[:created_at] = Time.now
      attrs[:ip_addresses] = JSON.parse(attrs[:ip_addresses]) rescue [] if attrs[:ip_addresses].is_a?(String)

      object = RedborderObject.find_or_initialize_by(id: attrs[:id])

      if object.new_record?
        object.assign_attributes(attrs)
        object.save
      else
        next if (!current_user.admin? && current_user.id != object.user_id)
        attrs.delete(:created_at)
        object.update(attrs)
      end
    end
  end

  #
  # Allowed params when create an object
  #
  def object_params
    params.require(:object).permit(:page, :search, :name, :value, :comment, :type, :user_id, :sensor_id, :vendor)
  end
end
