#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
# Author: Juan Sebastián Soto <jsoto@redborder.com>
#######################################################################

require 'net/http'
require 'uri'
require 'open-uri'
require 'digest/md5'
require 'yaml'

module Redborder
  module RulesUpdateSnort3

    def self.start(verbose=false, initial_tgz=nil, source_id=nil)

      @content = ""

      now = Time.now

      sources = source_id.nil? ? RuleSource.where(is_custom: false, enable: true) : RuleSource.where(id: source_id)

      @config = YAML.load_file("#{Rails.root}/config/rules2rbIPS.yml")

      logit "#{I18n.t 'lib_redborder.node'} " + `hostname`.chomp!

      logit ""

      logit "#{I18n.t 'lib_redborder.starting_rules'}"

      # Download rules from each source
      sources.each do |source|

        begin

          new_rules      = false    # if true we have new rules and we will create a new RuleDbVersion.
          download_files = false
          @warn_version   = false

          logit ""
          logit "############################################################################"
          logit "       #{I18n.t 'lib_redborder.rule_source'}: #{source.clean_name}"
          logit "############################################################################"

          if source.has_code and source.code.blank?
            # Print no code for source
            logit "INFO: #{I18n.t 'lib_redborder.the_source'} #{source.clean_name} #{I18n.t 'lib_redborder.needs'}."
            next
          end

          if source.is_custom
            tgz_tmp_file = "#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}"
            logit "#{I18n.t 'lib_redborder.downloading'} #{source.cached_tgz_path}" if verbose
            logit "  * #{I18n.t 'lib_redborder.saving'} #{tgz_tmp_file}" if verbose

            if StorageFile.s3_enabled?
              download_file(source.cached_tgz_uri.to_s, tgz_tmp_file)
            else
              FileUtils.copy_file(source.cached_tgz_uri.to_s, tgz_tmp_file)
            end

            if source.md5 != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
              new_rules = true
            else
              logit "  * #{I18n.t 'lib_redborder.the_upload'}." if verbose
            end
          else
            if source.cached_tgz.nil? # if no cached_tgz is stored, force download files
              download_files = true
            else
              md5_file = "#{@config["cache_dir"]}/#{source.clean_name}.md5"
              File.delete(md5_file) if File.exists?(md5_file)
              if source.md5url.present? # check downloaded md5sum with md5sum of current rule_db_version
                logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_md5url}" if verbose
                logit "  * #{I18n.t 'lib_redborder.saving'} #{md5_file}" if verbose

                download_file(source.full_md5url, md5_file)
                md5_file_content = File.open(md5_file) {|f| f.readline}.gsub(/^"/, '').chomp

                if md5_file_content != source.md5 # if md5sum is different, force download files
                  logit "     - Cached tgz MD5 -> #{source.md5}" if verbose
                  logit "     - #{source.full_tgzurl} MD5 -> #{md5_file_content}" if verbose
                  download_files = true
                else
                  logit "  * #{I18n.t 'lib_redborder.cached_file'}" if verbose
                end
              else
                download_files = true
              end
            end

            if download_files
              tgz_tmp_file = "#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}"

              logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_tgzurl}" if verbose
              logit "  * #{I18n.t 'lib_redborder.saving'} #{tgz_tmp_file}" if verbose

              download_file(source.full_tgzurl, tgz_tmp_file)

              #check if the file is correct
              if defined?(md5_file_content) and md5_file_content and md5_file_content != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
                logit "  * #{I18n.t 'lib_redborder.not_match'}" if verbose
                next # go to next source
              elsif defined?(md5_file_content) and md5_file_content
                new_rules = true
              elsif source.cached_tgz.nil? # source has no md5url and no cached_tgz
                new_rules = true
              elsif source.md5 != Digest::MD5.hexdigest(File.read(tgz_tmp_file))
                new_rules = true
              end
            end
          end

          next if !new_rules

          FileUtils.mv(tgz_tmp_file, "#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}")

          last_rule_db_version = source.rule_db_versions.where(completed: true).last
          rule_db_version = RuleDbVersion.create(rule_source: source, completed: false)
          rule_db_version.update(checked: true) if (last_rule_db_version.nil? or source.is_custom)

          parse_rules(source.id, rule_db_version.id, verbose)

          Object::Rule.transaction do |t|
            begin

              if last_rule_db_version.present?

                last_rule_db_version.rules.where(default_enabled: [true, false]).find_each do |r|
                  r.update(is_deprecated: true) if Object::Rule.where(rule_sid: r.rule_sid, rule_gid: r.rule_gid, rule_db_version: rule_db_version, default_enabled: [true, false]).last.nil?
                end

                rule_db_version.rules_deprecated_count = last_rule_db_version.rules.where(is_deprecated: true, default_enabled: [true, false]).count

              end

              source.cached_tgz = File.open("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}") unless source.is_custom
              source.md5 = Digest::MD5.hexdigest(File.read("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}"))
              source.save

              rule_db_version.completed              = true
              rule_db_version.rules_count            = rule_db_version.rules.where(default_enabled: [true, false]).count
              rule_db_version.rules_new_count        = rule_db_version.rules.where(is_new: true, default_enabled: [true, false]).count
              rule_db_version.rules_modified_count   = rule_db_version.rules.where(is_modified: true, default_enabled: [true, false]).count

              rule_db_version.save

            rescue Exception => e
              logit "#{I18n.t 'lib_redborder.an_error'} #{e.to_s}. #{I18n.t 'lib_redborder.rolling'}"
              raise ActiveRecord::Rollback
              return
            end
          end

        rescue Exception => e
          logit "#{I18n.t 'lib_redborder.error_source'} #{source.clean_name} (#{e})"
          e.backtrace.each do |b|
            logit b
          end

        ensure
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}-tmp.#{source.compression_type}")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}.md5")
          FileUtils.rm_rf("#{@config["cache_dir"]}/#{source.clean_name}")
        end

      end

      logit "\n#{I18n.t 'lib_redborder.warning'} #{Sensor::SNORT_VERSIONS.join(",")}\n" if @warn_version

      logit "#{I18n.t 'lib_redborder.exiting'}" if verbose

      return @content

    rescue Exception => e
      logit "#{I18n.t 'lib_redborder.error_updating'} (ERROR: #{e})"
      return @content
    end

    def self.logit(msg)
      @content ||= ""
      @content << "#{msg}\n"
    end    

    # Checks if new rule files exist by comparing cached and remote MD5 hashes.
    def self.check_if_new_rules?(source, tmp_tgz, final_tgz, verbose)
      new_rules = false
    
      logit "       - #{I18n.t('lib_redborder.checking_new_rules', source: source.clean_name)}"
      
      if source.cached_tgz.nil?
        logit "       - #{I18n.t('lib_redborder.no_cached_tgz')}"
        new_rules = true
      else
        logit "       - #{I18n.t('lib_redborder.has_cached_tgz')}"
        if source.md5url.present?
          logit "       - #{I18n.t('lib_redborder.md5_local', md5: source.md5)}"
          cache_dir = File.dirname(tmp_tgz)
          md5_file  = File.join(cache_dir, "#{source.clean_name}.md5")
          File.delete(md5_file) if File.exist?(md5_file)
    
          download_file(source.md5url, md5_file)
          md5_remote = File.read(md5_file).strip.gsub(/^"/, '').gsub(/"$/, '')
    
          logit "       - #{I18n.t('lib_redborder.md5_remote', md5: md5_remote)}"
          new_rules = true if md5_remote != source.md5
        else
          new_rules = true
        end
      end
    
      if new_rules
        download_file(source.tgzurl, tmp_tgz)
    
        if source.md5url.present?
          md5_file_content = File.read(tmp_tgz)
          local_md5 = Digest::MD5.hexdigest(md5_file_content)
          cache_dir = File.dirname(tmp_tgz)
          md5_file  = File.join(cache_dir, "#{source.clean_name}.md5")
    
          if File.exist?(md5_file)
            md5_remote = File.read(md5_file).strip.gsub(/^"/, '').gsub(/"$/, '')
            return false if md5_remote != local_md5
          end
        end
    
        FileUtils.mv(tmp_tgz, final_tgz)
      end
    
      new_rules
    end
    
    
    # Downloads a remote file from a given URL and saves it to the specified destination path.
    def self.download_file(url, destination)
      URI.open(url,
        ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE,
        read_timeout: 10
      ) do |remote_file|
        File.open(destination, "wb") do |file|
          file.write(remote_file.read)
        end
      end
    end
    

    def self.parse_rule_line(line)
      # Fix the line to remove any trailing comments
      match_data = line.strip.match(/\A(.*?)\s*\((.*)\)\s*\z/)
      return nil unless match_data

      {
        header: match_data[1].strip,
        body:   match_data[2].strip 
      }
    end

    def self.parse_header(header)
      tokens = header.split(/\s+/)
      disabled = (tokens.first == '#')
      tokens.shift if disabled

      action = tokens[0]

      # Por defecto, no hay protocolo, IPs, ni puertos
      protocol = src_ip = src_port = direction = dst_ip = dst_port = nil

      # If the rule has a protocol, IPs, and ports
      if tokens.size >= 7
        protocol   = tokens[1]
        src_ip     = tokens[2]
        src_port   = tokens[3]
        direction  = tokens[4]
        dst_ip     = tokens[5]
        dst_port   = tokens[6]
      end

      {
        disabled:     disabled,
        action:       action,
        protocol:     protocol,
        source_addr:  src_ip,
        source_port:  src_port,
        direction:    direction,
        target_addr:  dst_ip,
        target_port:  dst_port
      }
    end

    def self.parse_body(body)
      {
        sid:               body[/sid:\s*(\d+)/, 1]&.to_i,
        msg:               body[/msg:\s*"([^"]+)"/, 1],
        gid:               body[/gid:\s*(\d+)/, 1]&.to_i || 1,
        rev:               body[/rev:\s*(\d+)/, 1]&.to_i || 1,
        detection_filter:  body[/detection_filter:\s*([^;]+)/, 1],
        mitre_tactic_id:   body[/mitre_tactic_id\s+(TA\d+)/, 1] || "N/A",
        mitre_technique_id: body[/mitre_technique_id\s+(T\d+)/, 1] || "N/A"
      }
    end


    def self.parse_rules_file(file, source, rule_db_version, last_rule_db_version, verbose)
      category1 = File.basename(file, '.rules').strip.downcase
      category1 = category1.gsub(/#{source.regexp_group_ignore}/, '') if source.regexp_group_ignore

      logit "* Procesando archivo: #{File.basename(file)}"
      logit "  * #{I18n.t 'lib_redborder.category'}: #{category1}"

      # Agrupación por configuración de category1_groups
      category1_tmp = category1
      if @config['category1_groups']
        @config['category1_groups'].each do |catgr|
          if Regexp.new(catgr['regexp']).match?(category1)
            category1_tmp = catgr['name']
            break
          end
        end
      end

      return if category1_tmp == 'deleted'
      return unless File.exist?(file)

      # Variables auxiliares para categorización superior
      category_base  = category1.sub(/^snort3-/, '')   # → browser-chrome, file-image, etc.
      category_group = category1_tmp                  # → puede ser "browser-chrome" o grupo configurado

      # Guardar o reutilizar Group
      rc1 = Group.where(name: category1_tmp).first_or_create
      logit "  → Group: #{rc1.name}"

      # Categoría nivel 4 desde configuración
      category4 = "others"
      category4_desc = nil
      if @config['category4_groups']
        @config['category4_groups'].each do |catgr|
          if Regexp.new(catgr['regexp']).match?(category_base) || Regexp.new(catgr['regexp']).match?(category_group)
            category4 = catgr['name']
            category4_desc = catgr['desc']
            break
          end
        end
      end

      rc4 = Category.where(name: category4).first_or_create
      rc4.update(description: category4_desc) if category4_desc
      logit "  → Category: #{rc4.name} (#{rc4.description})"
      logit "This is rc4: #{rc4.inspect}"
      File.foreach(file) do |line|
        line = line.strip
        next if line.empty? || line.start_with?('#')
    
        parsed = parse_rule_line(line)
        next unless parsed

        header_data = parse_header(parsed[:header])
        body_data   = parse_body(parsed[:body])
        next unless body_data[:sid] && body_data[:msg]
    
        short_msg = begin
          body_data[:msg].gsub(/^#{category1.upcase} /, '')
        rescue
          body_data[:msg]
        end
    
        r = Object::Rule.new(
          rule_sid:           body_data[:sid],
          rule_gid:           body_data[:gid],
          rule_rev:           body_data[:rev],
          msg:                body_data[:msg],
          short_msg:          short_msg,
          mitre_tactic_id:    body_data[:mitre_tactic_id],
          mitre_technique_id: body_data[:mitre_technique_id],
          rule_db_version_id: rule_db_version.id,
          default_enabled:    !header_data[:disabled],
          protocol:           header_data[:protocol],
          source_addr:        header_data[:source_addr],
          source_port:        header_data[:source_port],
          target_addr:        header_data[:target_addr],
          target_port:        header_data[:target_port],
          text:               parsed[:body],
          group_id:           rc1.id,
          category_id:        rc4.id,
          snort_version:      RuleSource::S_VERSION_SNORT3
        )
    
        # Check if the rule is new or modified
        if last_rule_db_version
          prev = Object::Rule.where(
            rule_sid: r.rule_sid,
            rule_gid: r.rule_gid,
            rule_db_version_id: last_rule_db_version.id
          ).last
          r.is_new      = prev.nil?
          r.is_modified = !r.is_new && r.rule_rev != prev.rule_rev
        else
          r.is_new      = true
          r.is_modified = false
        end

        if r.save
          metadata_str = parsed[:body][/metadata:\s*([^;]+)/, 1]
          if metadata_str
            metadata_str.split(/,\s*/).each do |md|
              # policy
              if (m = /^policy\s+(?<name>\S+)\s+(?<action>\S+)/.match(md))
                policy = FactoryPolicy.where(name: m[:name]).first_or_create
                action = Action.where(name: m[:action]).first
                FactoryPolicyRule.create(rule_id: r.id, factory_policy_id: policy.id, action_id: action.id) if action
              # service
              elsif (m = /^service\s+(?<svc>\S+)/.match(md))
                svc = Service.where(name: m[:svc]).first_or_create
                svc.rules << r
              end
            end
          end
    
          # REFERENCES
          parsed[:body].scan(/reference:\s*([^,;]+),\s*([^;]+)/).each do |name, url|
            ref = RbReference.where(name: name).first_or_create(humanize: name)
            RbReferenceRule.create(rule_id: r.id, rb_reference_id: ref.id, url: url)
          end
    
          # FLOWBITS
          parsed[:body].scan(/flowbits:\s*([^,; ]+)(?:,\s*([^,; ]+))?(?:,\s*([^,; ]+))?/).each do |action, state, group|
            flowbit = Flowbit.where(name: action).first_or_create
            state_obj = state && FlowbitState.where(name: state).first_or_create
            group_obj = group && FlowbitGroup.where(name: group).first_or_create
    
            FlowbitRule.create(
              rule_id: r.id,
              flowbit_id: flowbit.id,
              flowbit_state: state_obj,
              flowbit_group: group_obj
            )
          end
        else
          logit "#{I18n.t 'lib_redborder.rule_not_saved'} #{line}; #{I18n.t 'lib_redborder.errors'} #{r.errors.inspect}" if verbose
        end
      end
    end
        
    def self.parse_rules(source_id, rule_db_version_id, verbose)
      source = RuleSource.find(source_id)
      last_rule_db_version = source.rule_db_versions.where(completed: true).last


      logit "#{I18n.t 'lib_redborder.new_rules'}" if verbose
      tgzLocalFile = "#{@config["cache_dir"]}/#{source.clean_name}.#{source.compression_type}"

      
    
      logit "#{I18n.t 'lib_redborder.extracting'} #{tgzLocalFile}" if verbose
      dir = "#{@config['cache_dir']}/#{source.clean_name}"
    
      FileUtils.rm_rf(dir)
      FileUtils.mkdir_p dir
    
      case source.compression_type
      when 'zip'
        system('unzip', tgzLocalFile, '-d', dir)
      when 'tgz'
        system('tar', 'xzf', tgzLocalFile, '-C', dir)
      end
    
      rule_db_version = RuleDbVersion.find(rule_db_version_id)
    
      # This is optional in snort3, but we will create a directory for it
      # and create a fake file to avoid errors
      so_rules_dir  = "#{@config['so_rules_dir']}/dbversion-#{rule_db_version.id}"
      FileUtils.mkpath so_rules_dir
    
      logit "#{I18n.t 'lib_redborder.processing'} #{source.clean_name} so_rules" if verbose
    
      so_rules = (source.so_rules || ['so_rules/precompiled/RHEL-6-0/x86-64'])
      so_rules.each do |rulesDir|
        logit "  * #{I18n.t 'lib_redborder.directory'}: #{rulesDir}" if verbose
        if !Dir.exist?("#{dir}/#{rulesDir}")
          logit "    - #{I18n.t 'lib_redborder.directory'} #{dir}/#{rulesDir} #{I18n.t 'lib_redborder.ignored'}" if verbose
        else
          Dir.open("#{dir}/#{rulesDir}") do |rulesDirectory|
            count = 0
            total = rulesDirectory.to_a.delete_if { |x| ['..', '.'].include? x }.size
    
            rulesDirectory.to_a.delete_if { |x| ['..', '.'].include? x }.each do |file|
              count += 1
              next if file == '.' || file == '..'

              # Skip if the file is not a .so file
              @warn_version = true unless Sensor::SNORT_VERSIONS.include?(file)
    
              logit "       - Reading #{file} ( #{(count*100.0/total).round(2)}% )" if verbose
              `cd #{dir}/#{rulesDir}/#{file} && if [ -f #{so_rules_dir}/snort-so_rules-#{file}.tar ]; then tar rf #{so_rules_dir}/snort-so_rules-#{file}.tar .; else tar cf #{so_rules_dir}/snort-so_rules-#{file}.tar .; fi`
            end
          end
        end
      end
    
      
      fake_so_rules = "#{so_rules_dir}/snort-so_rules.tar"
      system("rm -f /tmp/.fake; touch /tmp/.fake; tar cf #{fake_so_rules} -C /tmp .fake &>/dev/null; rm -f /tmp/.fake")
      system("gzip -q #{fake_so_rules};")
    
      # compress the so_rules
      Dir.open(so_rules_dir) do |so_rulesCompressed|
        so_rulesCompressed.to_a.each do |file|
          next if file == '.' || file == '..'
          next unless file.end_with?('tar')
    
          `gzip -q #{so_rules_dir}/#{file};`
    
          so_rule_match = /^snort-so_rules-(?<version>[\d\.]+).tar$/.match(file)
          next if so_rule_match.nil?
    
          so_versions = so_rule_match[:version].split('.')
          if so_versions.length > 2
            `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0]}.#{so_versions[1]}.tar.gz ] &&
               ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0]}.#{so_versions[1]}.tar.gz`
          end
    
          if so_versions.size > 3 && so_versions[3].to_i > 0
            `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.')}.tar.gz ] &&
               ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.')}.tar.gz`
    
            (0..so_versions[3].to_i - 1).each do |i|
              `[ ! -f #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.') + '.' + i.to_s}.tar.gz ] &&
                 ln -s #{fake_so_rules}.gz #{so_rules_dir}/snort-so_rules-#{so_versions[0..2].join('.') + '.' + i.to_s}.tar.gz`
            end
          end
        end
      end
    
     
      system("cd #{so_rules_dir}; tar czf #{@config["cache_dir"]}/#{source.clean_name}/so_rules.tgz *")
      rule_db_version.update(so_rules: File.open("#{@config["cache_dir"]}/#{source.clean_name}/so_rules.tgz"))
      FileUtils.rm_rf(so_rules_dir)
    
    
      logit "#{I18n.t 'lib_redborder.processing'} #{source.clean_name} #{I18n.t 'lib_redborder.rules'}" if verbose
      
      base_dir = "#{@config["cache_dir"]}/#{source.clean_name}"

      
      rules_paths = if source.search.present?
        source.search.map { |subdir| "#{base_dir}/#{subdir}" }.select { |path| Dir.exist?(path) }
      else
        [base_dir]
      end

      # If there are no valid paths in source.search, use recursive search
      if rules_paths.empty?
        rules_files = Dir.glob("#{base_dir}/**/*.rules")
      else
        rules_files = rules_paths.flat_map { |dir| Dir.glob("#{dir}/**/*.rules") }
      end

      # Process each rules file
      Object::Rule.transaction do
        begin
          rules_files.each_with_index do |file_path, index|
            logit "  * #{I18n.t 'lib_redborder.processing'} #{file_path} (#{index + 1}/#{rules_files.size})" 
            parse_rules_file(file_path, source, rule_db_version, last_rule_db_version, verbose)
          end
        rescue StandardError => e
          logit "#{I18n.t 'lib_redborder.an_error'} #{e}. #{I18n.t 'lib_redborder.rolling'}" if verbose
          raise ActiveRecord::Rollback
        end
      end

    end
    
  end
end
